package worldModel;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Properties;

import rescuecore2.standard.entities.Edge;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import utilities.Logger;
import agent.ActionCommandException;
import agent.PoliceForceAgent;

import com.infomatiq.jsi.IntProcedure;
import com.infomatiq.jsi.Rectangle;
import com.infomatiq.jsi.SpatialIndex;
import com.infomatiq.jsi.rtree.RTree;
import com.vividsolutions.jts.algorithm.Angle;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineSegment;

public class TrafficSimulator {
	private WorldModel model = null;
	public int agentRad = 500;
	public static final int BUFFER_PREC = 1;
	public int aRAD = agentRad;

	private ArrayList<Geometry> shapes = new ArrayList<Geometry>();
	private ArrayList<Geometry> shapesBeforeBuffer = new ArrayList<Geometry>();
	public Geometry sorroundShape = null;
	private HashMap<Integer, Blockade> blockades = new HashMap<Integer, Blockade>();
	private HashMap<Blockade, Rectangle> rectangles = new HashMap<Blockade, Rectangle>();
	private SpatialIndex blocksIndex = new RTree();
	public GeometryFactory geom = new GeometryFactory();

	public Geometry myShape = null;
	private Logger logger = null;

	public TrafficSimulator(WorldModel model, Logger logger) {
		this.logger = logger;
		this.model = model;
		blocksIndex = new RTree();
		blocksIndex.init(new Properties());
		resetConnectivity();
		geom = new GeometryFactory();
		blockades = new HashMap<Integer, Blockade>();
		rectangles = new HashMap<Blockade, Rectangle>();
	}

	private void log(String msg) {
		// logger.log(msg);
	}

	public void precomputeWorks() {
		makeShapes();
		updateAreas();
		setNeighbours();
		resetConnectivity();
	}

	public void clearBlockades() {
		blocksIndex = new RTree();
		blocksIndex.init(new Properties());
		blockades.clear();
	}

	public void resetConnectivity() {
		long l = System.nanoTime();
		for (Area area : model.areas) {
			for (int i = 0; i < area.entrances.size(); i++) {
				Entrance e1 = area.entrances.get(i);
				e1.internalEntrances = new ArrayList<Entrance>();
				e1.isItConnectedToNeighbour = true;
				e1.internalEntrances.addAll(area.entrances);
				e1.internalEntrances.remove(e1);
				// for (Entrance e2 : area.entrances)
				// if (e1 != e2)
				// e1.internalEntrances.add(e2);
			}
		}
		long m = System.nanoTime();
		long d = (m - l) / 1000000;

	}

	public void setNeighbours() {
		long l = System.nanoTime();
		for (Area area : model.areas) {
			for (Edge edge : area.serverArea.getEdges()) {
				if (edge.neighbourEdge != null && edge.entrance == null) {
					LineSegment ls = match(edge, area);
					if (ls != null) {
						LineSegment lsNeighbour = match(edge.neighbourEdge, edge.neighbourEdge.area.area);
						if (lsNeighbour != null) {
							Entrance entrance = new Entrance(area, ls, geom, area.entrances.size());
							Entrance neighbour = new Entrance(edge.neighbourEdge.area.area, lsNeighbour, geom,
									edge.neighbourEdge.area.area.entrances.size());
							entrance.neighbour = neighbour;
							neighbour.neighbour = entrance;
							entrance.area.entrances.add(entrance);
							neighbour.area.entrances.add(neighbour);

							edge.entrance = entrance;
							edge.neighbourEdge.entrance = neighbour;

							// for (int i = 0; i < 2; i++)
							// area.matchedCoords.add(ls.getCoordinate(i));
							// for (int i = 0; i < 2; i++)
							// neighbour.area.matchedCoords.add(lsNeighbour
							// .getCoordinate(i));
						}
					}
				}
			}
		}
		long m = System.nanoTime();
		long d = (m - l) / 1000000;

	}

	private LineSegment match(Edge edge, Area area) {
		LineSegment edgeSegment = new LineSegment(new Coordinate(edge.getStartX(), edge.getStartY()),
				new Coordinate(edge.getEndX(), edge.getEndY()));
		double edgeAngle = edgeSegment.angle();
		LineSegment bestMatch = null;
		double minDist = Integer.MAX_VALUE;
		for (int i = 0; i < area.shape.getCoordinates().length - 1; i++) {
			LineSegment ls = new LineSegment(area.shape.getCoordinates()[i], area.shape.getCoordinates()[i + 1]);

			double angle = ls.angle();
			double diff = Math.abs(Angle.diff(edgeAngle, angle));
			if (diff < 6 * Math.PI / 180 || diff > 174 * Math.PI / 180) {
				double mainDist = edgeSegment.distance(ls.midPoint());
				double dist = mainDist + Math.abs(diff) * 180.0 / Math.PI;
				if (mainDist < 40 && dist < minDist) {
					minDist = dist;
					bestMatch = ls;
				}
			}
		}

		return bestMatch;
	}

	@SuppressWarnings("unused")
	private String printLine(rescuecore2.misc.geometry.Line2D line) {
		LineSegment ls = new LineSegment(line.getOrigin().getX(), line.getOrigin().getY(), line.getEndPoint().getX(),
				line.getEndPoint().getY());
		return GeometryTools.printLine(ls);
	}

	public void updateAreas() {
		long l = System.nanoTime();
		if (model.simAgent.preCompute) {
			for (Geometry shape : shapes) {
				Rectangle rect = GeometryTools.makeRectangle(shape);
				for (Entity entity : model.model.getObjectsInRectangle((int) rect.min[0], (int) rect.min[1],
						(int) rect.max[0], (int) rect.max[1])) {
					if (entity instanceof rescuecore2.standard.entities.Area) {
						Area area = ((rescuecore2.standard.entities.Area) entity).area;
						area.circumstance = area.circumstance.difference(shape);
						if (area.circumstance.getArea() == 0) {
							aRAD -= 50;
							for (Area a : model.areas)
								a.circumstance = GeometryTools.createPolygonFromApexes(a.serverArea.getApexList());

							System.err.println("Bad Area: " + area.id);
							long m = System.nanoTime();
							long d = (m - l) / 1000000;
							bufferShapes();
							updateAreas();
							return;
						}
					}
				}
			}
			// TODO : mage too if e model.simAgent.precompute nistim ????
			if (model.simAgent.preCompute)
				for (Area area : model.areas) {
					area.circumstance = area.circumstance.intersection(sorroundShape);
					if (area.circumstance.getArea() == 0) {
						aRAD -= 50;
						for (Area a : model.areas)
							a.circumstance = GeometryTools.createPolygonFromApexes(a.serverArea.getApexList());

						System.err.println("Bad Area: " + area.id);
						long m = System.nanoTime();
						long d = (m - l) / 1000000;
						bufferShapes();
						updateAreas();
						return;
					}
				}

		}
		// TODO

		for (Area area : model.areas) {
			area.shape = area.circumstance;// .buffer(-10, BUFFER_PREC);
			// if (area.shape.getLength() == 0)
			// area.shape = area.circumstance;
		}
		long m = System.nanoTime();
		long d = (m - l) / 1000000;
	}

	private Coordinate changeSameCoords(Coordinate c, int indexOfC, int indexOfSame, ArrayList<Coordinate> coords) {
		Coordinate newCoordinate = new Coordinate(c.x - 1, c.y - 1);
		LineSegment newLineSegment1 = new LineSegment(coords.get(indexOfC - 1), newCoordinate);
		LineSegment newLineSegment2 = new LineSegment(coords.get(indexOfC + 1), newCoordinate);
		LineSegment lastLine1;
		if (indexOfSame != 0)
			lastLine1 = new LineSegment(coords.get(indexOfSame), coords.get(indexOfSame - 1));
		else
			lastLine1 = new LineSegment(coords.get(indexOfSame), coords.get(coords.size() - 2));
		LineSegment lastLine2 = new LineSegment(coords.get(indexOfSame), coords.get(indexOfSame + 1));
		if (newLineSegment1.intersection(lastLine1) != null || newLineSegment1.intersection(lastLine2) != null
				|| newLineSegment2.intersection(lastLine1) != null || newLineSegment2.intersection(lastLine2) != null) {
			newCoordinate.x += 2;
			newCoordinate.y += 2;
		}
		return newCoordinate;
	}

	public void makeShapes() {
		long l = System.nanoTime();
		HashSet<Edge> checkedEdges = new HashSet<Edge>();
		shapes = new ArrayList<>();
		shapesBeforeBuffer = new ArrayList<>();

		for (Area area : model.areas) {
			ArrayList<Coordinate> coords = new ArrayList<Coordinate>();

			Edge edge = null;
			for (Edge e : area.serverArea.getEdges()) {
				if (e.getNeighbour() == null && !checkedEdges.contains(e)) {
					checkedEdges.add(e);
					edge = e;
					break;
				}
			}

			if (edge == null)
				continue;

			boolean dir = true;
			coords.add(edge.getStart(dir));
			edge = edge.getNextEdge(dir);

			while (true) {
				while (edge.getNeighbour() == null && !checkedEdges.contains(edge)) {
					coords.add(edge.getStart(dir));
					checkedEdges.add(edge);
					edge = edge.getNextEdge(dir);
				}

				if (edge.getNeighbour() == null) {
					coords.add(edge.getStart(dir));
					break;
				}

				Coordinate lastCoord = edge.getStart(dir);
				while (edge.getNeighbour() != null) {
					edge = edge.neighbourEdge;
					dir = Math.hypot(edge.getStartX() - lastCoord.x, edge.getStartY() - lastCoord.y) > Math
							.hypot(edge.getEndX() - lastCoord.x, edge.getEndY() - lastCoord.y);
					edge = edge.getNextEdge(dir);
				}
			}
			try {
				if (coords.size() == 3) {
					LineSegment tmpLine = new LineSegment(coords.get(0), coords.get(1));
					coords.clear();
					Coordinate[] tmpLineCoords = tmpLine.toGeometry(geom).buffer(15).getCoordinates();
					for (Coordinate c : tmpLineCoords)
						coords.add(c);
				}
				shapes.add(GeometryTools.createPolygon(coords));
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		// fixShapes();
		shapesBeforeBuffer.addAll(shapes);
		long m = System.nanoTime();
		long d = (m - l) / 1000000;
		bufferShapes();
	}

	@SuppressWarnings("unused")
	private void fixShapes() {
		for (int i = 0; i < shapes.size(); i++) {
			Geometry p = shapes.get(i);
			for (int j = 0; j < p.getCoordinates().length - 1; j++)
				for (int k = j + 1; k < p.getCoordinates().length - 1; k++) {
					if (p.getCoordinates()[j].x == p.getCoordinates()[k].x
							&& p.getCoordinates()[j].y == p.getCoordinates()[k].y) {
						ArrayList<Coordinate> coords = new ArrayList<>();
						for (Coordinate c : p.getCoordinates())
							coords.add(c);
						Coordinate n = changeSameCoords(p.getCoordinates()[k], k, j, coords);
						coords.set(k, n);
						shapes.set(i, GeometryTools.createPolygon(coords));
						p = shapes.get(i);
					}
				}
		}
	}

	private void bufferShapes() {
		long l = System.nanoTime();
		Geometry testPoint = geom.createPoint(
				new Coordinate(model.areas.get(0).serverArea.getX(), model.areas.get(0).serverArea.getY()));
		for (int i = 0; i < shapes.size(); i++) {
			Geometry p = shapesBeforeBuffer.get(i);
			if (p.contains(testPoint)) {
				shapes.set(i, p.buffer(-aRAD + 30, BUFFER_PREC));
				sorroundShape = shapes.get(i);
			} else {
				Geometry pBuffered = p.buffer(aRAD - 30, BUFFER_PREC);
				int copyOfRad = aRAD;
				while (isBadShape(pBuffered)) {
					// System.out.println("BAD MAP:D :: "
					// + GeometryTools.printShape(pBuffered, 0));
					copyOfRad -= 50;
					pBuffered = p.buffer(copyOfRad - 30, BUFFER_PREC);
					if (copyOfRad <= 0) {
						pBuffered = p;
						break;
					}
				}
				if (copyOfRad < aRAD) {
					System.err.println("Bad map!");
					if (copyOfRad <= 0) {
						// System.out.println("shape: "
						// + GeometryTools.printShape(p, 0));
						System.err.println("AWFUL MAP!");
						shapes.remove(i);
						shapesBeforeBuffer.remove(i);
						i--;
						continue;
					}
				}
				shapes.set(i, pBuffered);
			}
		}
		shapes.remove(sorroundShape);
		long m = System.nanoTime();
		long d = (m - l) / 1000000;
	}

	private boolean isBadShape(Geometry s) {
		for (int j = 0; j < s.getCoordinates().length; j += 2)
			for (int k = j + 2; k < s.getCoordinates().length - 2; k += 2) {
				LineSegment l1 = new LineSegment(s.getCoordinates()[j], s.getCoordinates()[j + 1]);
				LineSegment l2 = new LineSegment(s.getCoordinates()[k], s.getCoordinates()[k + 1]);
				if (l1.intersection(l2) != null)
					return true;
			}
		return false;
	}

	// private void log(String msg) {
	// model.simAgent.log(msg);
	// }

	public HashSet<Entrance> update(ChangeSet changeSet, int myX, int myY) throws ActionCommandException {
		myShape = null;
		Geometry myPos = null;
		if (myX >= 0 && myY >= 0)
			myPos = geom.createPoint(new Coordinate(myX, myY));
		ArrayList<Area> areas = new ArrayList<Area>();
		for (EntityID entityID : changeSet.getChangedEntities()) {
			Entity entity = model.getEntity(entityID);
			if (entity instanceof rescuecore2.standard.entities.Area) {
				Area area = ((rescuecore2.standard.entities.Area) entity).area;

				for (Blockade blockade : area.blockades)
					removeBlockade(blockade);
				area.updateBlockades(model.model, agentRad);
				for (Blockade blockade : area.blockades) {
					addBlockade(blockade);
				}

				for (Entrance entrance : area.entrances)
					entrance.internalEntrances = new ArrayList<Entrance>();
				areas.add(area);
			}
		}
		for (Area area : areas)
			for (Entrance entrance : area.entrances)
				entrance.updated = false;
		for (Area area : areas) {
			updateAreaConnectivity(area);
		}
		HashSet<Entrance> reachableEntrances = new HashSet<Entrance>();
		for (Area area : areas) {
			for (Entrance e : area.entrances)
				e.shapeInArea = null;
			Geometry shape = area.circumstance.buffer(10);
			if (area.getID().getValue() == model.simAgent.getMyPos().id && !shape.contains(myPos))
				shape = area.completeShape;
			for (Blockade blockade : getBlockadesInRange(area)) {
				if (model.simAgent.hasRead)
					shape = shape.difference(blockade.circumstance);
				else
					shape = shape.difference(blockade.circumstance.buffer(agentRad, 1));
			}
			updateReachablityofArea(area, shape, myPos, reachableEntrances);
			try {
				if (agentRad == 500 && !(model.simAgent instanceof PoliceForceAgent)
						&& area.getID().getValue() == model.simAgent.getMyPos().id && reachableEntrances.size() == 0) {
					shape = area.circumstance.buffer(10);
					if (area.getID().getValue() == model.simAgent.getMyPos().id && !shape.contains(myPos)) {
						shape = area.completeShape;
					}
					Geometry temp = shape;
					done: for (int i = 10; i <= 20; i += 10) {
						for (Blockade blockade : getBlockadesInRange(area)) {
							try {
								shape = shape.difference(blockade.mainShape.buffer(i));
							} catch (Exception e) {
								if (i > 10) {
									if (model.simAgent.logger.getWriter() != null)
										e.printStackTrace(model.simAgent.logger.getWriter());
									e.printStackTrace();
								}
								shape = temp;
								continue done;
							}
						}
						break;
					}
					updateReachablityofArea(area, shape, myPos, reachableEntrances);
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
			area.areaWithOutBlock = shape;
		}
		return reachableEntrances;
	}

	public void updateAreaConnectivity(Area area) {
		for (Entrance entrance : area.entrances) {
			if (entrance.updated)
				continue;
			ArrayList<Blockade> blocks = getBlockadesInRange(GeometryTools.makeRectangle(entrance.lineSegment));
			entrance.updatePoints(blocks);
			entrance.neighbour.updatePoints(blocks);
			entrance.updateConnectivity(blocks, model.isInChangeSet(entrance.neighbour.area.getID()));
			entrance.updated = entrance.neighbour.updated = true;
		}
	}

	private void updateReachablityofArea(Area area, Geometry shape, Geometry myPos,
			HashSet<Entrance> reachableEntrances) {
		@SuppressWarnings("unchecked")
		ArrayList<Entrance> allEntrances = (ArrayList<Entrance>) area.entrances.clone();
		ArrayList<Geometry> areaShapes = new ArrayList<>();
		for (int i = 0; i < shape.getNumGeometries(); i++)
			areaShapes.add(shape.getGeometryN(i));
		Collections.sort(areaShapes, new Comparator<Geometry>() {
			public int compare(Geometry g1, Geometry g2) {
				if (g1.getArea() > g2.getArea())
					return 1;
				else if (g1.getArea() < g2.getArea())
					return -1;
				return 0;
			};
		});
		// TODO The condition below, is written because of the bug that 2
		// entrances is connected to an entrance but they aren't connected
		// to each others! This will make connecting the discussed entrance to
		// agent's shape prior to other shapes.
		if (myPos != null && area.id == model.simAgent.getMyPos().id)
			for (Geometry s : areaShapes)
				if (s.contains(myPos)) {
					myShape = s;
					ArrayList<Geometry> tmp = new ArrayList<>();
					tmp.addAll(areaShapes);
					areaShapes.clear();
					areaShapes.add(myShape);
					for (Geometry g : tmp)
						if (g != myShape)
							areaShapes.add(g);
					break;
				}

		for (Geometry s : areaShapes) {
			ArrayList<Entrance> neighbours = new ArrayList<Entrance>();
			done: for (Entrance entrance : allEntrances) {
				for (Geometry point : entrance.getPoints()) {
					if (s.contains(point)) {
						neighbours.add(entrance);
						continue done;
					}
				}
			}

			boolean amIInShape = false;
			if (myPos != null && area.id == model.simAgent.getMyPos().id)
				amIInShape = s.contains(myPos);
			if (amIInShape)
				myShape = s;
			for (int j = 0; j < neighbours.size(); j++) {
				Entrance entrance = neighbours.get(j);
				entrance.shapeInArea = s;
				if (amIInShape) {
					reachableEntrances.add(entrance);
				}
				for (Entrance other : neighbours)
					if (!entrance.equals(other))
						entrance.internalEntrances.add(other);
			}
			// TODO The condition below, is written because of the bug that 2
			// entrances is connected to an entrance but they aren't connected
			// to each others!
			if (neighbours.size() > 1)
				allEntrances.removeAll(neighbours);
		}
	}

	private ArrayList<Blockade> getBlockadesInRange(Area area) {
		return getBlockadesInRange(GeometryTools.makeRectangle(area.circumstance));
	}

	public ArrayList<Blockade> getBlockadesInRange(Rectangle rect) {
		final ArrayList<Blockade> blocks = new ArrayList<Blockade>();
		if (blocksIndex == null)
			System.out.println("blocksIndex is null");
		blocksIndex.intersects(rect, new IntProcedure() {
			public boolean execute(int id) {
				blocks.add(blockades.get(id));
				return true;
			}
		});
		return blocks;
	}

	private void addBlockade(Blockade blockade) {
		//TODO: mainShape Or circumstance?
		blockades.put(blockade.getID().getValue(), blockade);
		Rectangle rectangle = GeometryTools.makeRectangle(blockade.circumstance);
		rectangles.put(blockade, rectangle);
		blocksIndex.add(rectangle, blockade.getID().getValue());
	}

	public void removeBlockade(Blockade blockade) {
		blockades.remove(blockade.getID().getValue());
		if (blocksIndex == null)
			System.out.println("is null is null");
		blocksIndex.delete(rectangles.get(blockade), blockade.getID().getValue());
		rectangles.remove(blockade);
	}

}
