package worldModel;

import java.awt.geom.Rectangle2D;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map.Entry;
import java.util.Stack;
import java.util.TreeSet;

import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import agent.ActionCommandException;
import agent.Agent;
import agent.Agent.MapType;

import com.infomatiq.jsi.Rectangle;
import com.vividsolutions.jts.geom.Coordinate;

public class WorldModel implements Serializable {
	private static final long serialVersionUID = -5590912657021792186L;

	public transient ArrayList<Area> areas = new ArrayList<Area>();
	public ArrayList<Building> buildings = new ArrayList<Building>();
	public ArrayList<Road> roads = new ArrayList<Road>();
	public transient ArrayList<Refuge> refuges = new ArrayList<Refuge>();
	public transient ArrayList<Hydrant> hydrants = new ArrayList<Hydrant>();
	public transient ArrayList<Entrance> allEntrancesOfAreas = null;

	public transient HashSet<Entrance> myEntrances = new HashSet<Entrance>();
	public transient ArrayList<Area> reachableAreas = null;
	public transient ArrayList<Building> reachableBuildings = null;
	public transient ArrayList<Road> reachableRoads = null;
	public transient ArrayList<Refuge> reachableRefuges = null;
	public transient ArrayList<Hydrant> reachableHydrants = null;

	public transient Agent<? extends StandardEntity> simAgent = null;
	protected transient StandardWorldModel model;
	public transient ChangeSet changeSet = null;
	public transient ArrayList<PoliceForce> policeForces = new ArrayList<PoliceForce>();
	public transient ArrayList<AmbulanceTeam> ambulanceTeams = new ArrayList<AmbulanceTeam>();
	public transient ArrayList<FireBrigade> fireBrigades = new ArrayList<FireBrigade>();
	public transient ArrayList<AmbulanceCentre> ambulanceCenters = new ArrayList<AmbulanceCentre>();
	public transient ArrayList<PoliceOffice> policeOffices = new ArrayList<PoliceOffice>();
	public transient ArrayList<FireStation> fireStations = new ArrayList<FireStation>();
	public transient ArrayList<Human> agents = new ArrayList<Human>();

	public transient HashMap<Area, Integer> areaToIndex = null;
	public transient HashMap<Building, Integer> buildingToIndex = null;
	public transient HashMap<worldModel.Road, Integer> roadToIndex = null;
	// public ArrayList<Area> highWayAreas = new ArrayList<Area>();
	public transient ArrayList<Area> CopyOfAreas = new ArrayList<Area>();

	public int maxDistanceView = 30000;
	public int maxDistanceExtinguish = 0;

	protected transient TrafficSimulator trafficSimulator = null;

	int limitAStarBuildings = 10, limitAStarLayer = 10;

	public MapType map = MapType.Unknown;

	public WorldModel(Agent<? extends StandardEntity> simAgent,
			StandardWorldModel model) {
		this.simAgent = simAgent;
		this.model = model;
		setEdges();

		for (Entity e : model.getAllEntities())
			if (e instanceof rescuecore2.standard.entities.Building) {
				worldModel.Building b = null;
				if (e instanceof rescuecore2.standard.entities.Refuge) {
					Refuge r = new Refuge(
							(rescuecore2.standard.entities.Refuge) e);
					b = r;
					refuges.add(r);
				} else {
					b = new worldModel.Building(
							(rescuecore2.standard.entities.Building) e);
				}
				buildings.add(b);
				areas.add(b);
				// b.initializeWalls(this);
			} else if (e instanceof rescuecore2.standard.entities.Road) {
				Road r = null;
				if (e instanceof rescuecore2.standard.entities.Hydrant) {
					Hydrant h = new Hydrant(
							(rescuecore2.standard.entities.Hydrant) e);
					r = h;
					hydrants.add(h);
				} else {
					r = new Road((rescuecore2.standard.entities.Road) e);
				}
				roads.add(r);
				areas.add(r);
			}
		trafficSimulator = new TrafficSimulator(this, simAgent.logger);
		trafficSimulator.precomputeWorks();
		fillAllEntrances();

		fillAreaComparator();
		fillIDComparator();

		Collections.sort(areas, areaComparator);
		Collections.sort(buildings, areaComparator);
		Collections.sort(roads, areaComparator);
		Collections.sort(hydrants, areaComparator);
		Collections.sort(refuges, areaComparator);

		initializeModels();
		maxDistanceExtinguish = simAgent.getConfig().getIntValue(
				Agent.MAX_DISTANCE_EXTINGUISH_KEY);
		maxDistanceView = simAgent.getConfig().getIntValue(
				Agent.MAX_DISTANCE_VIEW_KEY);

		mapCalculations();
		fillingIndexesForRadar();
		fillingNears();
		// TODO use highway weight in search
		// initHighWay();
		// TODO
		initBFSDist();
		fillingAreasInRange();
		for (Area a : areas)
			a.isEmpty = new boolean[simAgent.cvRecognizer];
	}

	public void spitOfMexico() {
		int[] badAreas = { 50708, 117890, 119082, 119077, 94128, 117138,
				117668, 117669 };
		try {
			for (int i : badAreas)
				if (getAreaByInt(i) != null)
					getAreaByInt(i).canBeVisited = false;
		} catch (NullPointerException np) {
			return;
		}
	}

	private void fillingAreasInRange() {
		// int a = 0;
		for (Building build : buildings)
			for (StandardEntity se : getAreasInRange(build, setRange())) {
				// a++;
				if (areaToIndex
						.get(((rescuecore2.standard.entities.Area) se).area) != null)
					build.areasInRange
							.add(areaToIndex
									.get(((rescuecore2.standard.entities.Area) se).area));
			}
	}

	public int setRange() {
		int y = 0;
		if (map == MapType.Unknown)
			y = 35000;
		else if (map == MapType.Kobe)
			y = 30000;
		else if (map == MapType.VC)
			y = 30000;
		else if (map == MapType.Test)
			y = 10000;
		else if (map == MapType.Joao)
			y = 40000;
		else if (map == MapType.Eindhoven)
			y = 140000;
		else if (map == MapType.Berlin)
			y = 150000;
		else if (map == MapType.Paris)
			y = 35000;
		else if (map == MapType.Istanbul)
			y = 60000;
		else if (map == MapType.Mecxico)
			y = 60000;
		else if (map == MapType.NY)
			y = 86000;
		if (simAgent.noCommi)
			y *= 2;
		return y;
	}

	public ArrayList<StandardEntity> getAreasInRange(Area area, int range) {
		ArrayList<StandardEntity> areasInRange = new ArrayList<>();
		done: for (StandardEntity se : model.getObjectsInRange(area.getID(),
				range)) {
			boolean checkDist = false;
			if (se instanceof rescuecore2.standard.entities.Area
					&& se.getID() != area.getID()) {
				endCoords: for (Coordinate c1 : area.shape.getCoordinates()) {
					for (Coordinate c2 : ((rescuecore2.standard.entities.Area) se).area.shape
							.getCoordinates()) {
						double test = Math.hypot(c1.x - c2.x, c1.y - c2.y);
						if (test <= range) {
							checkDist = true;
							break endCoords;
						}
					}
				}
				if (!checkDist)
					continue done;
				areasInRange.add(se);
			}
		}
		return areasInRange;
	}

	public Entrance getEntranceToNeighbour(Area source, Area neighbour) {
		for (Entrance e : source.entrances)
			if (e.neighbour.area.id == neighbour.id)
				return e;
		return null;
	}

	public transient Comparator<Area> areaComparatorBasedOnCounter = new Comparator<Area>() {
		public int compare(Area a1, Area a2) {
			if (a1.counterForFindingHighways < a2.counterForFindingHighways)
				return 1;
			if (a1.counterForFindingHighways > a2.counterForFindingHighways)
				return -1;
			return 0;
		}
	};
	int minDist = 5000, maxForChoosingFirst = 30;

	// @SuppressWarnings("unused")
	// private void initHighWay() {
	// System.err.println("entrances " + allEntrancesOfAreas.size());
	// long l = System.nanoTime();
	// HashSet<Area> set = new HashSet<Area>();
	// Random r = new Random(324452);
	// Area first;
	// int dist;
	// ArrayList<Area> path;
	// done: for (int i = 0; i < maxForChoosingFirst; i++) {
	// first = buildings.get(r.nextInt(buildings.size()));
	// if (set.contains(first)) {
	// i--;
	// continue done;
	// }
	// for (Area area : set) {
	// dist = (int) Math.hypot((area.getX() - first.getX()),
	// (area.getY() - first.getY()));
	//
	// if (dist < minDist) {
	// i--;
	// continue done;
	// }
	// }
	// set.add(first);
	// for (Area area : buildings) {
	// if (area.id == first.id)
	// continue;
	// path = getMinPath(first.entrances, area, false);
	// for (Area a : path) {
	// if (a.id == first.id || a.id == area.id)
	// continue;
	// a.counterForFindingHighways++;
	// }
	// }
	// }
	// CopyOfAreas.addAll(areas);
	// Collections.sort(CopyOfAreas, areaComparatorBasedOnCounter);
	// String s = new String();
	// for (int i = 0; i < CopyOfAreas.size(); i++) {
	// if (i == allEntrancesOfAreas.size() / 32)
	// break;
	// highWayAreas.add(CopyOfAreas.get(i));
	// }
	// String str = new String();
	// for (Area area : set)
	// str += area.id + " ";
	// System.out.println("set " + str);
	// System.out.println("areas " + s);
	// long m = System.nanoTime();
	// long d = (m - l) / 1000000;
	// System.out.println("time of highway " + d);
	// }

	void mapCalculations() {
		simAgent.fillMapDetails();
		findBorders();
		map = mapDetection();
		if (map == MapType.Kobe || map == MapType.VC || map == MapType.Test)
			simAgent.bigMap = false;
		System.out.println("Map type: " + map);
	}

	public MapType mapDetection() {
		int[][] res = new int[5][5];
		done: for (Coordinate c : getTrafficSimulator().sorroundShape
				.convexHull().buffer(-30000).getCoordinates()) {
			for (int i = 0; i < res.length; i++)
				for (int j = 0; j < res[0].length; j++) {
					if (c.x > simAgent.mapBorders[i][j][0]
							&& c.x <= simAgent.mapBorders[i][j][2]
							&& c.y > simAgent.mapBorders[i][j][1]
							&& c.y <= simAgent.mapBorders[i][j][3]) {
						res[i][j]++;
						continue done;
					}
				}
		}

		for (Entry<MapType, int[][]> e : simAgent.mapDetails.entrySet())
			if (simAgent.diff(res, e.getValue()) <= 7)
				return e.getKey();
		return MapType.Unknown;
	}

	void findBorders() {
		double minX = Integer.MAX_VALUE, minY = Integer.MAX_VALUE, maxX = Double.MIN_VALUE, maxY = Double.MIN_VALUE;
		for (Area a : areas) {
			if (a.getX() < minX)
				minX = a.getX();
			if (a.getX() > maxX)
				maxX = a.getX();
			if (a.getY() < minY)
				minY = a.getY();
			if (a.getY() > maxY)
				maxY = a.getY();
		}
		double XStep = (maxX - minX) / simAgent.mapBorders.length;
		double YStep = (maxY - minY) / simAgent.mapBorders[0].length;
		for (int i = 0; i < simAgent.mapBorders.length; i++)
			for (int j = 0; j < simAgent.mapBorders[0].length; j++) {
				double[] a = new double[4];
				a[0] = minX + XStep * i;
				a[1] = minY + YStep * j;
				a[2] = minX + XStep * (i + 1);
				a[3] = minY + YStep * (j + 1);
				simAgent.mapBorders[i][j] = a;
			}
	}

	private void fillAllEntrances() {
		allEntrancesOfAreas = new ArrayList<Entrance>();
		for (Area area : areas)
			for (Entrance entrance : area.entrances)
				allEntrancesOfAreas.add(entrance);
	}

	private double setDistance(Area building) {
		double distance = 0;
		double max = 1;
		for (int i = 0; i < building.shape.getCoordinates().length; i++) {
			distance = Math.hypot(
					building.getX() - building.shape.getCoordinates()[i].x,
					building.getY() - building.shape.getCoordinates()[i].y);
			if (distance > max)
				max = distance;
		}
		return max;

	}

	private void fillingNears() {
		int x = 8000;
		// TODO
		if (!simAgent.bigMap)
			x = 200;
		for (Area area : areas) {
			if (area instanceof worldModel.Building) {
				int dist = Math.max(maxDistanceView, (int) setDistance(area)
						+ x);
				do {
					area.nearRoads.clear();
					for (StandardEntity e : getAreasInRange(area, dist)) {
						if (e instanceof rescuecore2.standard.entities.Road) {
							Road road = (Road) (((rescuecore2.standard.entities.Road) e).area);
							area.nearRoads.add(road);
						} else if (e instanceof rescuecore2.standard.entities.Building) {
							Building eB = (Building) (((rescuecore2.standard.entities.Building) e).area);
							if (!area.nearBuildings.contains(eB)) {
								area.nearBuildings.add(eB);
								if (area instanceof Building
										&& !eB.nearBuildings
												.contains((Building) area)) {
									eB.nearBuildings.add((Building) area);
								}
							}
						}
					}
					dist += 5000;
				} while (area.nearBuildings.size() <= 5);
			} else {
				for (StandardEntity e : getAreasInRange(area, maxDistanceView)) {
					if (e.getID().getValue() != area.getID().getValue()) {
						if (e instanceof rescuecore2.standard.entities.Area) {
							if (e instanceof rescuecore2.standard.entities.Road) {
								Road road = (Road) (((rescuecore2.standard.entities.Road) e).area);
								area.nearRoads.add(road);
							} else if (e instanceof rescuecore2.standard.entities.Building) {
								Building building = (Building) (((rescuecore2.standard.entities.Building) e).area);
								area.nearBuildings.add(building);
							}
						}
					}
				}
			}
			for (StandardEntity e : getAreasInRange(area, maxDistanceExtinguish)) {
				if (e instanceof rescuecore2.standard.entities.Area
						&& e.getID().getValue() != area.getID().getValue()) {
					area.nearAreas50000
							.add(((rescuecore2.standard.entities.Area) e).area);
				}
			}
		}
		for (Area area : areas) {
			Collections.sort(area.nearRoads, areaComparator);
			Collections.sort(area.nearBuildings, areaComparator);
			area.nearAreas.addAll(area.nearRoads);
			area.nearAreas.addAll(area.nearBuildings);
			if (area instanceof Building)
				area.nearBuildingsIncludeSelf.add((Building) area);
			else if (area instanceof Road)
				area.nearRoadsIncludeSelf.add((Road) area);
			area.nearBuildingsIncludeSelf.addAll(area.nearBuildings);
			area.nearRoadsIncludeSelf.addAll(area.nearRoads);
			area.nearAreasIncludeSelf.addAll(area.nearRoadsIncludeSelf);
			area.nearAreasIncludeSelf.addAll(area.nearBuildingsIncludeSelf);
			area.nearRoadsInts.clear();
			area.nearBuildingsInts.clear();
			for (Road r : area.nearRoads)
				area.nearRoadsInts.add(roadToIndex.get(r));
			for (Building b : area.nearBuildings)
				area.nearBuildingsInts.add(buildingToIndex.get(b));
		}
	}

	public int calculateDist(Entrance e1, Entrance e2) {
		if (e1.area instanceof Road)
			return (int) Math.hypot(e1.center.x - e2.center.x, e1.center.y
					- e2.center.y);
		else if (e1.area instanceof worldModel.Building && e1.area == e2.area)
			return 0;
		else if (e1.area instanceof worldModel.Building)
			return (int) Math.hypot(e1.center.x - e2.center.x, e1.center.y
					- e2.center.y) / 2;

		return 0;

	}

	public void bfs(Area area) {
		/* worldModel. */clearAreas();
		ArrayList<Entrance> layer = new ArrayList<Entrance>();
		layer.addAll(area.entrances);
		for (Entrance e : layer)
			e.mark = true;
		area.bfsDist = 0;
		int dist = 1;
		while (layer.size() > 0) {
			ArrayList<Entrance> newLayer = new ArrayList<Entrance>();
			for (Entrance enterance : layer) {
				if (!enterance.neighbour.mark) {
					enterance.neighbour.area.bfsDist = dist;
					enterance.neighbour.mark = true;
					for (Entrance internal : enterance.neighbour.area.entrances)
						if (!internal.mark) {
							newLayer.add(internal);
							internal.mark = true;
						}
				}
			}
			dist++;
			layer = newLayer;
		}
	}

	public double[][] BFSDist = null;

	public void initBFSDist() {
		BFSDist = new double[policeForces.size()][agents.size()];
		for (int i = 0; i < policeForces.size(); i++) {
			bfs(getArea(policeForces.get(i).getPosition()));
			for (int j = 0; j < agents.size(); j++) {
				BFSDist[i][j] = getArea(agents.get(j).getPosition()).bfsDist;
			}
		}
		// System.out.println("BFS Time " + d);
	}

	public transient Comparator<Area> areaComparator = null;
	{
		fillAreaComparator();
	}

	public void clearAreas() {
		for (Entrance enterance : allEntrancesOfAreas) {
			enterance.mark = false;
			enterance.lastEntrance = null;
			enterance.area.checkSituation4Fire = false;
		}
	}

	public void fillAreaComparator() {
		areaComparator = new Comparator<Area>() {
			public int compare(Area a1, Area a2) {
				if (a1.getID().getValue() > a2.getID().getValue())
					return 1;
				if (a1.getID().getValue() < a2.getID().getValue())
					return -1;
				return 0;
			}
		};
	}

	public void clearTargets() {
		for (worldModel.Area area : areas)
			area.setAreaAsTarget(false);
	}

	public transient Comparator<Entity> IDComparator = null;
	{
		fillIDComparator();
	}

	public void fillIDComparator() {
		IDComparator = new Comparator<Entity>() {
			public int compare(Entity e1, Entity e2) {
				if (e1.getID().getValue() > e2.getID().getValue())
					return 1;
				if (e1.getID().getValue() < e2.getID().getValue())
					return -1;
				return 0;
			}
		};
	}

	public void newings() {
		policeForces = new ArrayList<PoliceForce>();
		ambulanceTeams = new ArrayList<AmbulanceTeam>();
		fireBrigades = new ArrayList<FireBrigade>();
		agents = new ArrayList<Human>();
		policeOffices = new ArrayList<PoliceOffice>();
		ambulanceCenters = new ArrayList<AmbulanceCentre>();
		fireStations = new ArrayList<FireStation>();
		myEntrances = new HashSet<>();
		reachableAreas = new ArrayList<>();
		reachableBuildings = new ArrayList<>();
		reachableHydrants = new ArrayList<>();
		reachableRefuges = new ArrayList<>();
		reachableRoads = new ArrayList<>();

		CopyOfAreas = new ArrayList<Area>();
	}

	public void setTransients(Agent<? extends StandardEntity> simAgent,
			StandardWorldModel model) {
		this.simAgent = simAgent;
		this.model = model;
		areas = new ArrayList<>();
		refuges = new ArrayList<>();
		hydrants = new ArrayList<>();

		areas.addAll(buildings);
		areas.addAll(roads);
		Collections.sort(areas, areaComparator);

		for (Area area : areas) {
			rescuecore2.standard.entities.Area serverArea = (rescuecore2.standard.entities.Area) getEntityByInt(area.id);
			area.serverArea = serverArea;
			serverArea.area = area;
			area.fireMarked = new ArrayList<>();
			area.CVmarked = new HashSet<>();
			if (area instanceof Refuge)
				refuges.add((Refuge) area);
			if (area instanceof Hydrant)
				hydrants.add((Hydrant) area);
		}
		fillAllEntrances();

		newings();

		fillingIndexesForRadar();

		fillAreaComparator();
		fillIDComparator();

		initializeModels();

		trafficSimulator = new TrafficSimulator(this, simAgent.logger);

		for (Area a : areas)
			a.setTransients(this, trafficSimulator.geom);
		// setEdges();
		if (map == MapType.Kobe || map == MapType.VC || map == MapType.Joao
				|| map == MapType.Test)
			simAgent.bigMap = false;

	}

	public void fillingIndexesForRadar() {
		areaToIndex = new HashMap<>();
		buildingToIndex = new HashMap<>();
		roadToIndex = new HashMap<>();
		int iA = 0, iR = 0, iB = 0;
		for (Area a : areas) {
			areaToIndex.put(a, iA);
			iA++;
			if (a instanceof Building) {
				buildingToIndex.put((Building) a, iB);
				iB++;
			} else {
				roadToIndex.put((Road) a, iR);
				iR++;
			}
		}
	}

	public ArrayList<Civilian> getCivilians() {
		ArrayList<Civilian> civilians = new ArrayList<Civilian>();
		for (StandardEntity sa : model
				.getEntitiesOfType(StandardEntityURN.CIVILIAN))
			civilians.add((Civilian) sa);
		return civilians;
	}

	protected void initializeModels() {
		for (Entity e : model.getAllEntities()) {
			Human h = null;
			if (e instanceof AmbulanceCentre)
				ambulanceCenters.add((AmbulanceCentre) e);
			else if (e instanceof PoliceOffice)
				policeOffices.add((PoliceOffice) e);
			else if (e instanceof FireStation)
				fireStations.add((FireStation) e);
			if (!(e instanceof Human))
				continue;
			h = (Human) e;
			if (!(h instanceof Civilian))
				agents.add(h);
			if (h instanceof PoliceForce)
				policeForces.add((PoliceForce) h);
			else if (h instanceof AmbulanceTeam)
				ambulanceTeams.add((AmbulanceTeam) h);
			else if (h instanceof FireBrigade)
				fireBrigades.add((FireBrigade) h);
		}
		Collections.sort(agents, IDComparator);
		Collections.sort(policeForces, IDComparator);
		Collections.sort(fireBrigades, IDComparator);
		Collections.sort(ambulanceTeams, IDComparator);
		Collections.sort(ambulanceCenters, IDComparator);
		Collections.sort(policeOffices, IDComparator);
		Collections.sort(fireStations, IDComparator);
	}

	public void update(ChangeSet changeSet) throws ActionCommandException {
		this.changeSet = changeSet;
		myEntrances = trafficSimulator.update(changeSet, simAgent.getX(),
				simAgent.getY());
		// updateReachableAreas();
	}

	public void updateReachableAreas() {
		reachableAreas = new ArrayList<Area>();
		reachableBuildings = new ArrayList<worldModel.Building>();
		reachableRoads = new ArrayList<Road>();
		reachableRefuges = new ArrayList<Refuge>();
		reachableHydrants = new ArrayList<Hydrant>();
		for (Area area : areas) {
			if (area instanceof Building)
				((Building) area).distToFireCalculated = false;
			area.distanceFromSelf = -1;
			area.isReachable = false;
			area.buildingReachabilityCalculated = false;
			area.buildingReachability = false;
		}
		for (Entrance e : allEntrancesOfAreas) {
			e.isReachable = false;
			e.marked = false;
			e.distanceFromSelf = -1;
		}

		int dist = 0;
		ArrayList<Entrance> layer = new ArrayList<Entrance>();
		layer.addAll(myEntrances);
		Collections.sort(layer, new Comparator<Entrance>() {

			@Override
			public int compare(Entrance e1, Entrance e2) {
				if (e1.id > e2.id)
					return 1;
				else if (e2.id > e1.id)
					return -1;
				return 0;
			}
		});

		while (layer.size() > 0) {
			ArrayList<Entrance> newLayer = new ArrayList<Entrance>();
			for (Entrance e : layer) {
				if (e.marked)
					continue;
				addReachableArea(e, dist);
				if (e.isItConnectedToNeighbour) {
					Entrance neighbour = e.neighbour;
					if (!neighbour.marked) {
						addReachableArea(neighbour, dist + 1);
						newLayer.addAll(neighbour.internalEntrances);
					}
				}
			}

			layer = newLayer;
			dist++;
		}
	}

	private void addReachableArea(Entrance enterance, int dist) {
		enterance.marked = true;
		enterance.isReachable = true;
		enterance.distanceFromSelf = dist;
		Area area = enterance.area;
		if (!area.isReachable) {
			area.isReachable = true;
			area.distanceFromSelf = dist;
			reachableAreas.add(area);
			if (area instanceof worldModel.Building)
				reachableBuildings.add((worldModel.Building) area);
			if (area instanceof Road) {
				reachableRoads.add((Road) area);
			}
			if (area instanceof Refuge)
				reachableRefuges.add((Refuge) area);
			if (area instanceof Hydrant)
				reachableHydrants.add((Hydrant) area);
		}
	}

	protected void setEdges() {

		for (Entity entity : model.getAllEntities()) {
			if (entity instanceof rescuecore2.standard.entities.Area) {
				rescuecore2.standard.entities.Area area = (rescuecore2.standard.entities.Area) entity;

				// set nextEdge
				for (int i = 1; i < area.getEdges().size(); i++) {
					area.getEdges().get(i - 1).nextEdge = area.getEdges()
							.get(i);
					area.getEdges().get(i).prevEdge = area.getEdges()
							.get(i - 1);
				}
				area.getEdges().get(area.getEdges().size() - 1).nextEdge = area
						.getEdges().get(0);
				area.getEdges().get(0).prevEdge = area.getEdges().get(
						area.getEdges().size() - 1);

				// set area
				for (Edge edge : area.getEdges()) {
					edge.area = area;
				}

				// set neighbour
				for (Edge edge : area.getEdges()) {
					if (edge.getNeighbour() != null
							&& edge.neighbourEdge == null) {
						int xMid = (edge.getStartX() + edge.getEndX()) / 2;
						int yMid = (edge.getStartY() + edge.getEndY()) / 2;

						Edge neighbour = null;
						double minDist = Integer.MAX_VALUE;
						for (Edge e : ((rescuecore2.standard.entities.Area) model
								.getEntity(edge.getNeighbour())).getEdges()) {
							if (entity.getID().equals(e.getNeighbour())) {
								int x = (e.getStartX() + e.getEndX()) / 2;
								int y = (e.getStartY() + e.getEndY()) / 2;
								double dist = Math.hypot(x - xMid, y - yMid);
								if (dist < minDist) {
									neighbour = e;
									minDist = dist;
								}
							}
						}

						if (neighbour != null) {
							edge.neighbourEdge = neighbour;
							neighbour.neighbourEdge = edge;
						}
					}
				}
			}
		}
	}

	public ArrayList<Area> getMinPath(Collection<Entrance> start, Area end) {
		return getMinPath(start, end, true);
	}

	public ArrayList<Area> getMinPath(Collection<Entrance> start, Area end,
			boolean checkBlockades) {
		setAreaAsTarget(end);
		return getMinPath(start, checkBlockades);
	}

	public void setAreaAsTarget(Area end) {
		for (Entrance entrance : allEntrancesOfAreas)
			entrance.isTarget = false;
		for (Entrance entrance : end.entrances)
			entrance.isTarget = true;
	}

	public ArrayList<Area> getMinPath(Area start, Area end) {
		return getMinPath(start, end, true);
	}

	public ArrayList<Area> getMinPath(Area start, Area end,
			boolean checkBlockades) {
		setAreaAsTarget(end);
		return getMinPath(start, checkBlockades);
	}

	public ArrayList<Area> getMinPath(Area start) {
		return getMinPath(start, true);
	}

	public ArrayList<Area> getMinPath(Area start, boolean checkBlockades) {
		return getMinPath(start.entrances, checkBlockades);
	}

	public ArrayList<Area> getMinPath(Collection<Entrance> start) {
		return getMinPath(start, true);
	}

	public ArrayList<Area> getMinPath(Collection<Entrance> start,
			boolean checkBlockades) {

		for (Entrance entrance : allEntrancesOfAreas) {
			entrance.marked = false;
			entrance.lastEntrance = null;
		}
		ArrayList<Area> path = new ArrayList<Area>();
		ArrayList<Entrance> layer = new ArrayList<Entrance>();
		for (Entrance enterance : start) {
			if (enterance.isTarget) {

				return path;
			} else if (!checkBlockades || !enterance.marked) {
				enterance.marked = true;
				layer.add(enterance);
			}
		}
		Entrance lastEntrance = null;
		done: while (layer.size() > 0) {
			ArrayList<Entrance> newLayer = new ArrayList<Entrance>();
			for (Entrance enterance : layer) {
				if ((!checkBlockades || enterance.isItConnectedToNeighbour)
						&& !enterance.neighbour.marked) {
					Entrance neighbour = enterance.neighbour;
					neighbour.lastEntrance = enterance;
					if (neighbour.isTarget) {
						lastEntrance = neighbour;
						break done;
					}

					neighbour.marked = true;
					if (checkBlockades) {
						for (Entrance internal : neighbour.internalEntrances)
							if (!internal.marked) {
								internal.lastEntrance = neighbour;
								if (internal.isTarget) {
									lastEntrance = internal;
									break done;
								}

								internal.marked = true;
								newLayer.add(internal);
							}
					} else {
						for (Entrance internal : neighbour.area.entrances)
							if (neighbour != internal && !internal.marked) {
								internal.lastEntrance = neighbour;
								if (internal.isTarget) {
									lastEntrance = internal;
									break done;
								}

								internal.marked = true;
								newLayer.add(internal);
							}
					}
				}
			}

			layer = newLayer;
		}
		if (lastEntrance == null)
			return null;
		Stack<Area> stack = new Stack<Area>();
		while (lastEntrance != null) {
			Area area = lastEntrance.area;
			if (stack.size() == 0 || stack.peek() != area)
				stack.add(area);
			lastEntrance = lastEntrance.lastEntrance;
		}

		while (stack.size() != 0)
			path.add(stack.pop());
		return path;
	}

	public ArrayList<Area> getMinPathByHashSetfortail(
			Collection<Entrance> start, HashSet<Entrance> end,
			boolean checkBlockades) {

		ArrayList<Area> path = new ArrayList<Area>();
		HashSet<Entrance> mark = new HashSet<Entrance>();
		HashMap<Entrance, Entrance> lastEntrances = new HashMap<Entrance, Entrance>();
		HashSet<Entrance> targets = new HashSet<Entrance>();
		targets.addAll(end);
		ArrayList<Entrance> layer = new ArrayList<Entrance>();
		for (Entrance enterance : start) {
			mark.add(enterance);
			layer.add(enterance);
		}

		Entrance lastEntrance = null;
		done: while (layer.size() > 0) {
			// simAgent.log("mark : " + mark);
			ArrayList<Entrance> newLayer = new ArrayList<Entrance>();
			for (Entrance enterance : layer) {
				// simAgent.log("enterance: " + enterance);
				if ((!checkBlockades || enterance.isItConnectedToNeighbour)
						&& !mark.contains(enterance.neighbour)) {
					Entrance neighbour = enterance.neighbour;
					lastEntrances.put(neighbour, enterance);
					if (targets.contains(neighbour)) {
						// simAgent.log("we reached a target : " + neighbour);
						lastEntrance = neighbour;
						break done;
					}

					mark.add(neighbour);
					if (checkBlockades) {
						for (Entrance internal : neighbour.internalEntrances)
							if (!mark.contains(internal)) {
								lastEntrances.put(internal, neighbour);
								if (targets.contains(neighbour)) {
									lastEntrance = internal;
									break done;
								}

								mark.add(internal);
								newLayer.add(internal);
							}
					} else {
						for (Entrance internal : neighbour.area.entrances)
							if (neighbour != internal
									&& !mark.contains(internal)) {
								lastEntrances.put(internal, neighbour);
								if (targets.contains(internal)) {
									lastEntrance = internal;
									break done;
								}

								mark.add(internal);
								newLayer.add(internal);
							}
					}
				}
			}

			layer = newLayer;
		}
		if (lastEntrance == null) {
			return null;
		}

		Stack<Area> stack = new Stack<Area>();
		while (lastEntrance != null) {
			Area area = lastEntrance.area;
			if (stack.size() == 0 || stack.peek() != area)
				stack.add(area);
			lastEntrance = lastEntrances.get(lastEntrance);
		}

		while (stack.size() != 0)
			path.add(stack.pop());

		return path;
	}

	public ArrayList<Area> getMinPathByHashSet(Collection<Entrance> start,
			HashSet<Entrance> end, boolean checkBlockades) {
		ArrayList<Area> path = new ArrayList<Area>();
		for (Entrance e : start)
			if (end.contains(e))
				return path;

		return getMinPathByHashSetfortail(start, end, checkBlockades);
	}

	public ArrayList<Area> getMinPathByHashSet(Area start, Area end,
			boolean checkBlockades) {
		if (start == end)
			return new ArrayList<Area>();
		HashSet<Entrance> endEntrances = new HashSet<Entrance>();
		endEntrances.addAll(end.entrances);
		return getMinPathByHashSet(start.entrances, endEntrances,
				checkBlockades);

	}

	public ArrayList<EntityID> getMinPathDijkByHash(EntityID end,
			boolean checkBlockades, boolean isCalledForMove,
			WorldModel worldModel, Human myHum) {
		Area target = worldModel.getArea(end);
		Area a = worldModel.getArea(myHum.getPosition());
		if (a.equals(target)) {
			ArrayList<EntityID> path = new ArrayList<EntityID>();
			path.add(end);
			return path;
		}
		if (!isCalledForMove) {
			ArrayList<Area> pArea = getMinPath(a, target, checkBlockades);
			ArrayList<EntityID> p = new ArrayList<EntityID>();
			for (Area tofArea : pArea)
				p.add(tofArea.getID());
			if (p.size() > 30)
				end = p.get(20);
		}

		HashMap<Entrance, Entrance> lastEntrances = new HashMap<Entrance, Entrance>();
		final HashMap<Entrance, Double> dijkDist = new HashMap<Entrance, Double>();
		HashSet<Entrance> targets = new HashSet<Entrance>();

		TreeSet<Entrance> allEnterances = new TreeSet<Entrance>(
				new Comparator<Entrance>() {
					public int compare(Entrance o1, Entrance o2) {
						double d1 = 1000000;
						if (dijkDist.containsKey(o1))
							d1 = dijkDist.get(o1);
						double d2 = 1000000;
						if (dijkDist.containsKey(o2))
							d2 = dijkDist.get(o2);
						if (d1 > d2)
							return 1;
						if (d1 < d2)
							return -1;
						if (o1.area.getID().getValue() > o2.area.getID()
								.getValue())
							return 1;
						if (o1.area.getID().getValue() < o2.area.getID()
								.getValue())
							return -1;
						if (o1.id > o2.id)
							return 1;
						if (o1.id < o2.id)
							return -1;
						return 0;
					}
				});

		ArrayList<Entrance> firstEnterances = new ArrayList<Entrance>();
		firstEnterances.addAll(myEntrances);
		if (!checkBlockades)
			firstEnterances = a.entrances;
		for (Entrance e : firstEnterances) {
			Coordinate p2 = new Coordinate(a.getX(), a.getY());
			dijkDist.put(e, p2.distance(e.center));
			allEnterances.add(e);
		}
		// for (Area area : areas)
		// allEnterances.addAll(area.entrances);
		for (Entrance e : target.entrances)
			targets.add(e);

		while (allEnterances.size() > 0) {
			Entrance e = allEnterances.first();
			allEnterances.remove(e);
			ArrayList<Entrance> neighbours = new ArrayList<Entrance>();
			if (e.isItConnectedToNeighbour || !checkBlockades) {
				neighbours.add(e.neighbour);
			}
			if (checkBlockades)
				neighbours.addAll(e.internalEntrances);
			else
				for (Entrance enterance : e.area.entrances)
					if (enterance != e)
						neighbours.add(enterance);
			Coordinate ep = e.center;
			for (Entrance enter : neighbours) {
				Coordinate np = enter.center;
				double dist0 = ep.distance(np);
				if ((e.area.entrances.contains(enter) && !e.internalEntrances
						.contains(enter))
						|| (e.neighbour == enter && !enter.isItConnectedToNeighbour))
					dist0 += 5000;
				if (enter.area.isNearFire) {
					if (enter.area instanceof worldModel.Building
							&& ((worldModel.Building) enter.area).isOnFire()) {
						// System.out.println("1getMinPathDijk"+enter.area);
						dist0 *= 200;
					} else
						dist0 = 1;
				} else if (enter.area instanceof worldModel.Building
						&& ((worldModel.Building) enter.area).isOnFire()) {
					// System.out.println("2getMinPathDijk"+enter.area);
					dist0 *= 200;
				}
				if (!dijkDist.containsKey(enter)
						|| dijkDist.get(e) + dist0 < dijkDist.get(enter)) {
					// simAgent.log("3-Dijk"+enter.area);
					if (allEnterances.contains(enter))
						allEnterances.remove(enter);
					dijkDist.put(enter, dijkDist.get(e) + dist0);
					allEnterances.add(enter);
					lastEntrances.put(enter, e);
					if (targets.contains(enter)) {
						Entrance enterance = enter;
						ArrayList<EntityID> way1 = new ArrayList<EntityID>();
						ArrayList<EntityID> way2 = new ArrayList<EntityID>();
						while (enterance != null) {
							EntityID entityID = enterance.area.getID();
							if (way1.size() == 0
									|| way1.get(way1.size() - 1).getValue() != entityID
											.getValue())
								way1.add(entityID);
							if (lastEntrances.containsKey(enterance))
								enterance = lastEntrances.get(enterance);
							else
								enterance = null;
						}
						for (int i = way1.size() - 1; i >= 0; i--)
							way2.add(way1.get(i));
						return way2;
					}
				}
			}
		}
		return null;
	}

	@SuppressWarnings("unused")
	public ArrayList<EntityID> getMinPathDijk(EntityID end,
			boolean checkBlockades, boolean isCalledForMove,
			WorldModel worldModel, Human myHum) {
		// System.out.println("first");
		if (true)
			return getMinPathDijkByHash(end, checkBlockades, isCalledForMove,
					worldModel, myHum);
		// System.out.println("next");
		Area target = worldModel.getArea(end);
		TreeSet<Entrance> allEnterances = new TreeSet<Entrance>();
		Area a = worldModel.getArea(myHum.getPosition());
		if (a.equals(target)) {
			ArrayList<EntityID> path = new ArrayList<EntityID>();
			path.add(end);
			return path;
		}
		if (!isCalledForMove) {
			ArrayList<Area> pArea = getMinPath(a, target, checkBlockades);
			ArrayList<EntityID> p = new ArrayList<EntityID>();
			for (Area tofArea : pArea)
				p.add(tofArea.getID());
			if (p.size() > 30)
				end = p.get(20);
		}
		for (Area area : areas)
			for (Entrance ents : area.entrances) {
				ents.distForDijk = 1000000000;
				ents.prev = null;
				ents.isTarget = false;
			}

		ArrayList<Entrance> firstEnterances = new ArrayList<Entrance>();
		firstEnterances.addAll(myEntrances);
		if (!checkBlockades)
			firstEnterances = a.entrances;
		for (Entrance e : firstEnterances) {
			Coordinate p2 = new Coordinate(a.getX(), a.getY());
			e.distForDijk = p2.distance(e.center);
		}
		for (Area area : areas)
			allEnterances.addAll(area.entrances);
		for (Entrance e : target.entrances)
			e.isTarget = true;

		while (allEnterances.size() > 0) {
			Entrance e = allEnterances.first();
			allEnterances.remove(e);
			ArrayList<Entrance> neighbours = new ArrayList<Entrance>();
			if (e.isItConnectedToNeighbour || !checkBlockades) {
				neighbours.add(e.neighbour);
			}
			if (checkBlockades)
				neighbours.addAll(e.internalEntrances);
			else
				for (Entrance enterance : e.area.entrances)
					if (enterance != e)
						neighbours.add(enterance);
			Coordinate ep = e.center;
			for (Entrance enter : neighbours) {
				Coordinate np = enter.center;
				double dist0 = ep.distance(np);
				if ((e.area.entrances.contains(enter) && !e.internalEntrances
						.contains(enter))
						|| (e.neighbour == enter && !enter.isItConnectedToNeighbour))
					dist0 += 5000;
				// System.out.println("distForDijk" +enter.area);
				if (enter.area.isNearFire) {
					if (enter.area instanceof worldModel.Building
							&& ((worldModel.Building) enter.area).isOnFire()) {
						// System.out.println("if1:");
						dist0 *= 200;
					} else
						dist0 = 1;
				} else if (enter.area instanceof worldModel.Building
						&& ((worldModel.Building) enter.area).isOnFire()) {
					// System.out.println("if2:");
					dist0 *= 200;
				}
				if (e.distForDijk + dist0 < enter.distForDijk) {
					allEnterances.remove(enter);
					enter.distForDijk = e.distForDijk + dist0;
					allEnterances.add(enter);
					enter.prev = e;
					if (enter.isTarget) {
						Entrance enterance = enter;
						ArrayList<EntityID> way1 = new ArrayList<EntityID>();
						ArrayList<EntityID> way2 = new ArrayList<EntityID>();
						while (enterance != null) {
							EntityID entityID = enterance.area.getID();
							if (way1.size() == 0
									|| way1.get(way1.size() - 1).getValue() != entityID
											.getValue())
								way1.add(entityID);
							enterance = enterance.prev;
						}
						for (int i = way1.size() - 1; i >= 0; i--)
							way2.add(way1.get(i));
						return way2;
					}
				}
			}
		}
		return null;
	}

	public ArrayList<EntityID> getMinPathDijk(boolean checkBlockades,
			boolean isCalledForMove, WorldModel worldModel, Human myHum) {
		TreeSet<Entrance> allEnterances = new TreeSet<Entrance>();
		Area myPos = worldModel.getArea(myHum.getPosition());
		ArrayList<Entrance> firstEnterances = new ArrayList<Entrance>();
		firstEnterances.addAll(myEntrances);
		if (!checkBlockades)
			firstEnterances = myPos.entrances;
		for (Entrance e : firstEnterances) {
			Coordinate p2 = new Coordinate(myPos.getX(), myPos.getY());
			e.distForDijk = p2.distance(e.center);
		}

		for (Area area : areas)
			allEnterances.addAll(area.entrances);

		while (allEnterances.size() > 0) {
			Entrance e = allEnterances.first();
			allEnterances.remove(e);
			ArrayList<Entrance> neighbours = new ArrayList<Entrance>();
			if (e.isItConnectedToNeighbour || !checkBlockades) {
				neighbours.add(e.neighbour);
			}
			if (checkBlockades)
				neighbours.addAll(e.internalEntrances);
			else
				for (Entrance enterance : e.area.entrances)
					if (enterance != e)
						neighbours.add(enterance);
			Coordinate ep = e.center;
			for (Entrance enter : neighbours) {
				Coordinate np = enter.center;
				double dist0 = ep.distance(np);
				if (!e.internalEntrances.contains(enter)
						|| !enter.isItConnectedToNeighbour)
					dist0 += 5000;
				if (enter.area.isNearFire) {
					if (enter.area instanceof worldModel.Building
							&& ((worldModel.Building) enter.area).isOnFire())
						dist0 *= 200;
					else
						dist0 = 1;
				} else if (enter.area instanceof worldModel.Building
						&& ((worldModel.Building) enter.area).isOnFire())
					dist0 *= 200;
				if (e.distForDijk + dist0 < enter.distForDijk) {
					allEnterances.remove(enter);
					enter.distForDijk = e.distForDijk + dist0;
					allEnterances.add(enter);
					enter.prev = e;
					if (enter.isTarget) {
						Entrance enterance = enter;
						ArrayList<EntityID> way1 = new ArrayList<EntityID>();
						ArrayList<EntityID> way2 = new ArrayList<EntityID>();
						while (enterance != null) {
							EntityID entityID = enterance.area.getID();
							if (!way1.contains(entityID))
								way1.add(entityID);
							enterance = enterance.prev;
						}
						for (int i = way1.size() - 1; i >= 0; i--)
							way2.add(way1.get(i));
						return way2;
					}
				}
			}
		}
		return null;
	}

	public TrafficSimulator getTrafficSimulator() {
		return trafficSimulator;
	}

	public boolean isInChangeSet(EntityID id) {
		if (changeSet == null)
			return false;
		return changeSet.getChangedEntities().contains(id);
	}

	public Entity getEntity(EntityID id) {
		return model.getEntity(id);
	}

	public Entity getEntityByInt(int id) {
		return getEntity(new EntityID(id));
	}

	public Area getArea(EntityID id) {
		return ((rescuecore2.standard.entities.Area) getEntity(id)).area;
	}

	public Area getAreaByInt(int id) {
		return getArea(new EntityID(id));
	}

	public StandardWorldModel getModel() {
		return model;
	}

	private Rectangle2D bound;

	public Rectangle2D getBounds() {
		int minX = Integer.MAX_VALUE;
		int maxX = Integer.MIN_VALUE;
		int minY = Integer.MAX_VALUE;
		int maxY = Integer.MIN_VALUE;
		// Add all rectangles
		for (StandardEntity next : model) {
			Rectangle r = makeRectangle(next);
			if (r != null) {
				minX = Math.min(minX, (int) r.min[0]);
				maxX = Math.max(maxX, (int) r.max[0]);
				minY = Math.min(minY, (int) r.min[1]);
				maxY = Math.max(maxY, (int) r.max[1]);
			}
		}
		bound = new Rectangle2D.Double(minX, minY, maxX - minX, maxY - minY);

		return bound;
	}

	private Rectangle makeRectangle(StandardEntity e) {
		int x1 = Integer.MAX_VALUE;
		int x2 = Integer.MIN_VALUE;
		int y1 = Integer.MAX_VALUE;
		int y2 = Integer.MIN_VALUE;
		if (e instanceof rescuecore2.standard.entities.Area) {
			int[] apexes = ((rescuecore2.standard.entities.Area) e)
					.getApexList();
			if (apexes.length == 0) {
				return null;
			}
			for (int i = 0; i < apexes.length - 1; i += 2) {
				x1 = Math.min(x1, apexes[i]);
				x2 = Math.max(x2, apexes[i]);
				y1 = Math.min(y1, apexes[i + 1]);
				y2 = Math.max(y2, apexes[i + 1]);
			}
		} else if (e instanceof rescuecore2.standard.entities.Blockade) {
			int[] apexes = ((rescuecore2.standard.entities.Blockade) e)
					.getApexes();
			if (apexes.length == 0) {
				return null;
			}
			for (int i = 0; i < apexes.length - 1; i += 2) {
				x1 = Math.min(x1, apexes[i]);
				x2 = Math.max(x2, apexes[i]);
				y1 = Math.min(y1, apexes[i + 1]);
				y2 = Math.max(y2, apexes[i + 1]);
			}
		} else if (e instanceof Human) {
			Human h = (Human) e;
			rescuecore2.misc.geometry.Point2D location = h.getPositionPoint();
			if (location == null) {
				return null;
			}
			x1 = (int) location.getX();
			x2 = (int) location.getX();
			y1 = (int) location.getY();
			y2 = (int) location.getY();
		} else
			System.err.println("unknown entity passed to make rectangle");
		return new Rectangle(x1, y1, x2, y2);
	}

	boolean indexed = false;

}